/*! \file
    \brief Header file for TBDML routines

 */
#ifndef _TBDML_H_
#define _TBDML_H_

// Header file for TBDML.DLL
#if defined(DLL)
//! This definition is used when TBDML.h is being exported
#define TBDML_API WINAPI __declspec(dllexport)
#else
//! This definition is used when TBDML.h is being imported
#define TBDML_API WINAPI __declspec(dllimport)
#endif

//! MACRO to calculate a SYNC value from a given frequency in Hz
#define SYNC_MULTIPLE(x)  (U16)((2*128*(60000000UL/10))/(x/10))

//!  Target microcontroller types
typedef enum {
   HC12  = 0,     //!< HC12 or HCS12 target
   HCS08 = 1,     //!< HCS08 target
   RS08  = 2,     //!< RS08 target
   CFV1  = 3,     //!< Coldfire Version 1 target
   T_OFF = 0xFF,  //!< Turn off interface (no target)
} target_type_e;

//!  Reset mode (normal operation or BDM)
typedef enum { // Type of reset action required
   SPECIAL_MODE   = 0,  //!< Special mode (BDM active, Target halted)
   NORMAL_MODE    = 1   //!< Normal mode (usual reset, Target executes)
} target_mode_e;

//!  Target RS08 microcontroller derivatives
typedef enum {
  KA1 = 0, //!< RS08KA1
  KA2 = 1, //!< RS08KA2
  LA8 = 2, //!< RS08LA8
  KA4 = 4, //!< RS08KA4 (testing only)
  KA8 = 5, //!< RS08KA8 (testing only)
} rs08_derivative_type_e;

//! Target supports ACKN or use fixed delay (WAIT) instead
typedef enum {
   ACKN  = 1,   //!< Target supports ACKN feature and it is enabled
   WAIT  = 0,   //!<  Use WAIT (delay) instead
} ackn_e;

//! Target reset status values
typedef enum {
   NO_RESET_ACTIVITY    = 0,   //!< No reset activity since last polled
   RESET_DETECTED       = 1    //!< Reset since last polled
} reset_e;

//! Target speed selection
typedef enum {
   NO_INFO              = 0,   //!< Not connected
   SYNC_SUPPORTED       = 1,   //!< Speed determined by SYNC
   SPEED_GUESSED        = 2,   //!< Speed determined by trial & error
   SPEED_USER_SUPPLIED  = 3    //!< User has specified the speed to use
} speed_e;

//! Target status
typedef struct { // state of BDM communication
   ackn_e         ackn_state;       //!< Supports ACKN ?
   reset_e        reset_state;      //!< Reset recently ?
   speed_e        connection_state;
} bdm_status_t;

//! Target microcontroller registers (HCS12, HCS08/RS08)
typedef union {
   //! HCS12 Registers
   struct {
      unsigned int pc;  //!< Program counter
      unsigned int sp;  //!< Stack Pointer
      unsigned int ix;  //!< Index register X
      unsigned int iy;  //!< Index register Y
      unsigned int d;   //!< D accumulator (A:B)
      unsigned int ccr; //!< Condition code register
   } hc12;
   //! HCS08 Registers (also used for RS08)
   struct {
      unsigned int pc;  //!< Program Counter (HCS08), CCR+PC (RS08)
      unsigned int sp;  //!< Stack Pointer (HCS08), Shadow PC (RS08)
      unsigned int hx;  //!< H:X index register (HCS08 only)
      unsigned int a;   //!< Accumulator (HCS08 + RS08)
      unsigned int ccr; //!< Condition code register (HCS08 only)
   } hcs08;
} registers_t;

#if defined(DLL)
//! This definition is used when TBDML is being used as a DLL
BOOL TBDML_API DllMain(HINSTANCE hDLLInst, DWORD fdwReason, LPVOID lpvReserved);
#endif

TBDML_API unsigned char  _tbdml_dll_version(void); 				     /* returns DLL version in BCD */
TBDML_API unsigned char  _tbdml_init(void);						        /* initialises USB and returns number of devices found */
TBDML_API unsigned char  _tbdml_open(unsigned char device_no); 	  /* opens a device with given number (0...), returns 0 on success and 1 on error */
TBDML_API void           _tbdml_close(void);								  /* closes currently open device */
TBDML_API unsigned int   _tbdml_get_version(void);				        /* returns version of HW (MSB) and SW (LSB) of the interface in BCD format */
TBDML_API unsigned char  _tbdml_get_last_sts(void);				     /* returns status of the last command: 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_set_target_type(target_type_e target_type); /* sets target MCU type, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_target_sync(void); 				     /* measures BDM frequency of the target using the SYNC feature and connects to it, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_target_reset(target_mode_e target_mode); /* resets the target to normal or special mode, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_bdm_sts(bdm_status_t *bdm_status);  /* fills user supplied structure with actual state of BDM communication channel, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_read_bd(unsigned int address); 	  /* reads one byte from the BDM area of memory */
TBDML_API unsigned char  _tbdml_write_bd(unsigned int address, unsigned char data); /* writes one byte to the BDM area of memory, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_target_go(void);                    /* starts target execution from current PC address, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_target_step(void); 				     /* steps over a single target instruction, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_target_halt(void); 				     /* brings the target into active background mode, returns 0 on success and non-zero on failure */
TBDML_API unsigned char  _tbdml_set_speed(float crystal_frequency); /* sets the communication speed, parameter is crystal frequency in MHz, returns 0 on success and non-zero on failure */
TBDML_API float          _tbdml_get_speed(void); 						  /* returns crystal frequency of the target in MHz */
TBDML_API unsigned char  _tbdml_read_byte(unsigned int address);    /* reads one byte from memory */
TBDML_API void           _tbdml_write_byte(unsigned int address, unsigned char data); /* writes one byte to memory */
TBDML_API unsigned int   _tbdml_read_word(unsigned int address); 	  /* reads one word from memory (address must be aligned) */
TBDML_API void           _tbdml_write_word(unsigned int address, unsigned int data); /* writes one word to memory (address must be aligned) */
TBDML_API void           _tbdml_read_block(unsigned int address, unsigned int count, unsigned char *data); /* reads block from memory, data is returned to the user supplied buffer */
TBDML_API void           _tbdml_write_block(unsigned int address, unsigned int count, unsigned char *data); /* writes block to memory */
TBDML_API unsigned char  _tbdml_read_regs(registers_t *registers);      /* reads contents of core registers, returns 0 on success and non-zero on failure */
TBDML_API void           _tbdml_write_reg_pc(unsigned int value); 		/* writes PC */
TBDML_API void           _tbdml_write_reg_sp(unsigned int value); 		/* writes SP */
TBDML_API void           _tbdml_write_reg_x(unsigned int value);	 		/* writes H:X (HCS08) or IX (HC12) */
TBDML_API void           _tbdml_write_reg_y(unsigned int value); 			/* writes IY (HC12 only) */
TBDML_API void           _tbdml_write_reg_d(unsigned int value); 			/* writes B:A (HC12) or A (HCS08) */
TBDML_API void           _tbdml_write_reg_ccr(unsigned int value); 		/* writes CCR */

TBDML_API void           _usbdm_setBoot(void);

TBDML_API void           _usbdm_icp_reboot( void );
TBDML_API unsigned int   _usbdm_icp_program(unsigned int addr, unsigned int count, unsigned char *data);
TBDML_API unsigned int   _usbdm_icp_verify(unsigned int addr, unsigned int count, unsigned char *data);
TBDML_API unsigned int   _usbdm_icp_erase(unsigned int addr, unsigned int count);
TBDML_API unsigned int   _usbdm_icp_get_result( void );
#endif // _TBDML_H_
