/*! \file
    \brief Main command procedure for executing BDM commands.
    
    This file processes the commands received over the USB link from the host

   \verbatim
   *** Prominent Notice-This software was modified from TBDML software - 12/05 ***

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   \endverbatim

   \verbatim
   Change Hisory
   +====================================================================
   | 23 Jul 2008 | Added ICP code                                 - pgo
   | 13 Jul 2008 | Moved commandBuffer from Z_PAGE to make larger - pgo
   | 10 Jul 2008 | Changed re-connect strategy                    - pgo
   | 20 Jun 2008 | Fixed RS08 CMD_WRITE_8 when doing Flash prog.  - pgo
   | 20 Jun 2008 | Changed autoconnect condition                  - pgo
   | 13 Jun 2008 | Changed mem_write/read (coldfire)              - pgo
   |  8 Jun 2008 | Added CFv1 code                                - pgo
   | 11 Apr 2008 | Added reconnect code (SYNC on READ_STATUS)     - pgo
   | 11 Apr 2008 | Added option code                              - pgo
   | 23 Mar 2008 | Changed return code to use BDM_RC_... codes    - pgo
   |  3 Mar 2008 | Started changes for JM60 - lots                - pgo
   +====================================================================
   \endverbatim
*/
#include <string.h>
#include "derivative.h"
#include "Common.h"
#include "Configure.h"
#include "bdm.h"
#include "bdmMacros.h"
#include "commands.h"
#include "CmdProcessing.h"
#include "usb.h"
#include "icp.h"
#include "TargetDefines.h"
//#include "main.h"

#pragma MESSAGE DISABLE C4000 // Disable warnings about allways true
#pragma MESSAGE DISABLE C5900 // Disable warnings about result is zero

volatile static U8 (*deferredFunction)(void) = NULL;
/* buffer for Rx and Tx of commands & results */
/* data is always received starting at commandBuffer+1 and transmitted starting at commandBuffer+0 */
/* this is to make sure that status of the last command is preserved when receving the next command from the host */

//#pragma DATA_SEG __SHORT_SEG Z_PAGE
U8 commandBuffer[MAX_PACKET_SIZE];   
//#pragma DATA_SEG DEFAULT

//! Creates status byte
//!
//! @return 8-bit status byte made up as follows: \n
//!    - bit0 - ACKN is supported, \n
//!    - bit1 - Target was reset (this bit is cleared after reading),  \n
//!    - bit2 - Target has power \n
//!    - bit4-3 - Communication status: 00=NOT CONNECTED, 01=SYNC, 10=GUESS,  11=USER SUPPLIED
//!
U8 makeStatusByte(void) {
U8 status = 0;

   if (bdm_status.ackn==ACKN) // target supports ACKN and the feature is enabled ?
      status |= 0x01;                
#if (CAPABILITY&CAP_RESET)
   if (bdm_status.reset==RESET_DETECTED) {
      status |= 0x02;                // the target was resently reset externally
      if (RESET_IN!=0) 
         bdm_status.reset=NO_RESET_ACTIVITY;  // clear the flag if reset pin has gone high already
   }
#endif
   if (bdm_status.power)                      // target has power ?
      status |= 0x04;                
   if (bdm_status.speed==SPEED_USER_SUPPLIED) // do not blame me if it does not work...
      status |= 0x18; 
   else if (bdm_status.speed==SYNC_SUPPORTED) // SYNC feature used to detect speed
      status |= 0x08; 
   else if (bdm_status.speed==SPEED_GUESSED)  // speed guessed by trial & error
      status |=0x10; 
      
   return status;
}

//!  Processes all commands received over USB
//!
//!  @entry The command is expected to be in commandBuffer[1]
//!      
//!  @return Number of bytes left in commandBuffer to be sent back as response.\n
//!         commandBuffer[0] == commandBuffer[1] => Success \n
//!         commandBuffer[0] != commandBuffer[1] => Failure,  commandBuffer[0]  has the error code
//!   
static U8 _commandExec(void) {
#pragma DATA_SEG __SHORT_SEG Z_PAGE
static U8 rc;
static U8 lastCommand;
#pragma DATA_SEG DEFAULT

#ifdef DEBUG_PIN
   DEBUG_PIN_DDR = 1;
   DEBUG_PIN = 1;
   DEBUG_PIN = 0;
#endif   

   if (commandBuffer[1] == CMD_GET_LAST_STATUS) {
      // Return status from last command received
      if (commandBuffer[0] != lastCommand)
         commandBuffer[0] = BDM_RC_FAIL;
      else
         commandBuffer[0] = BDM_RC_OK;
      return(1);
   }

   lastCommand = commandBuffer[1];
   
   if (commandBuffer[0] == BDM_RC_BUSY) {
      // Don't accept further commands until last one has completed
      // Problems here with async thread!!!!!
      commandBuffer[0] = BDM_RC_OVERRUN;
      return(1);
   }

   commandBuffer[0] = commandBuffer[1]; // Assume the command will execute OK
   
   switch (commandBuffer[1]) {
#if (DEBUG&DEBUG_COMMANDS)
      // Various debugging & testing commands
      case CMD_DEBUG:
         switch (commandBuffer[2]) {
            case BDM_DBG_ACKN: // try the ACKN feature
               bdm_acknInit();  
               commandBuffer[1] = makeStatusByte(); // return the status byte
               return(2);    
            case BDM_DBG_SYNC: // try the sync feature
               rc = bdm_syncMeasure();  
               if (rc != BDM_RC_OK) {
                  commandBuffer[0] = rc;
                  return (1);
               }
               commandBuffer[1] = makeStatusByte(); // return the status byte
               return(2);    
            case BDM_DBG_TESTPORT: // Check port I/O timing - hangs USB interface
               bdm_checkTiming();
               return(1);
            case BDM_DBG_USBDISCONNECT: // CPU Reset (USB dis-connect)
               CTL = 0x00;      // Disable USB
               forceICPReset(); // Reset CPU - doesn't return!
#if (CAPABILITY & CAP_FLASH)
            case BDM_DBG_VPP_OFF: // Programming voltage off
               VPP_OFF();
               return(1);
            case BDM_DBG_VPP_ON:
               // Programming voltage on (requires FLASH12V on first)
               rc = bdmRS08_vppOn();
               if (rc != BDM_RC_OK) {
                  commandBuffer[0] = rc;
                  return (1);
               }
               return(1);
            case BDM_DBG_FLASH12V_OFF: // 12V charge pump off
               FLASH12V_OFF();
               return(1);
            case BDM_DBG_FLASH12V_ON: // 12V charge pump on
               FLASH12V_ON();
               return(1);
#endif // (CAPABILITY & CAP_FLASH)
            case BDM_DBG_VDD_OFF: // Target Vdd voltage off
               VDD_OFF();
               return(1);
            case BDM_DBG_VDD3_ON: // Target Vdd voltage on
               VDD3_ON();
               return(1);
            case BDM_DBG_VDD5_ON: // Target Vdd voltage on
               VDD5_ON();
               return(1);
            case BDM_DBG_CYCLE_POWER: // Cycle power to target
               commandBuffer[0] = bdm_cycleTargetVdd();
               return(1);
            case BDM_DBG_MEASURE_VDD: // Measure Target Vdd
               (*(U16*)(commandBuffer+2)) = bdm_targetVddMeasure(); // return the value
               commandBuffer[1] = commandBuffer[3];
               return(3);    
#if (DEBUG&STACK_DEBUG)
            case BDM_DBG_STACKSIZE: // Measure stack size
               {
               extern char far __SEG_START_SSTACK[];  // Bottom of stack space
               extern char far __SEG_END_SSTACK[];    // Top of stack space

               asm {
                  ldhx @__SEG_START_SSTACK // Bottom of stack space
               loop:
                  aix   #1
                  tst   0,x     // Find 1st used (non-zero) byte
                  beq   loop
                  
                  // Calculate size of stack in use
                  pshx
                  pshh
                  lda   @__SEG_END_SSTACK
                  sub   2,sp
                  sta   commandBuffer:1
                  lda   @__SEG_END_SSTACK:MSB
                  sbc   1,sp
                  sta   commandBuffer:2
                  ais   #2
                  }
               return(3);
               }
#endif // (DEBUG&STACK_DEBUG)
            case BDM_DBG_RS08TRIM:
               rc = bdmRS08_trimTargetClock(SYNC_MULTIPLE(8000000UL));
               if (rc != BDM_RC_OK) {
                  commandBuffer[0] = rc;
                  return (1);
               }
               commandBuffer[1] = savedTrimValue;
               return 2;
               break;
            case BDM_DBG_TESTWAITS:
               bdm_checkWaitTiming();
               return 1; // Never actually resturns.
            case BDM_DBG_TESTALTSPEED:
               return 1;
            default:
               commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
               return(1);    
         }
#endif // (DEBUG&DEBUG_COMMANDS)

      /* set/clear various options */
      case CMD_OPTION: {
         U8 option = commandBuffer[2];
         U8 value  = commandBuffer[3];
         switch (option) {
            case BDM_GET_CAPABILITIES : // Gets capability vector for hardware
               commandBuffer[1] = CAPABILITY; // returns 16-bit LittleEndian #
               commandBuffer[2] = 0;
               return (3);
               break;
            case BDM_OPT_SET_TARGET_VDD :       // Set target Vdd (off, 3V3 or 5V)
               if (value > BDM_TARGET_VDD_5V) {
                  commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS;
                  return (1);
               }
               bdm_option.targetVdd = value;
               break;
            case BDM_OPT_CYCLE_TARGET_VDD_ON_RESET : // Enable/disable Cycling Target Vdd on reset
               bdm_option.cycleVddOnReset = (value != 0);
               break;
            case BDM_OPT_CYCLE_TARGET_VDD_ON_CONNECT : // Enable/disable Cycling Target Vdd if connection problems
               bdm_option.cycleVddOnConnect = (value != 0);
               break;
            case BDM_OPT_AUTOMATIC_RECONNECT :  // Enable/disable automatic reconnect
               bdm_option.autoReconnect = (value != 0);
               break;
            case BDM_OPT_GUESS_SPEED :          // Enable/disable guessing speed for target w/o ACKN feature
               bdm_option.guessSpeed = (value != 0);
               break;
            case BDM_OPT_LEAVE_TARGET_POWERED : // Enable/disable leaving target power on exit
               bdm_option.leaveTargetPowered = (value != 0);
               break;
            case BDM_OPT_USE_ALT_BDM_CLOCK : // Enable/disable force BDM clock source in target
               bdm_option.useAltBDMClock = value;
               break;
            case BDM_OPT_USE_RESET_SIGNAL :  // Enable/disable use of RESET signal on BDM interface
               bdm_option.useResetSignal = (value != 0);
               break;
            case BDM_OPT_SET_CLOCK_TRIM_FREQ :
               bdm_option.targetClockTrim = (U16)((commandBuffer[4]<<8)+commandBuffer[3]);
               break;
            default :
               commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS;
               return (1);
         }
         return (1);
      }

      /* get HW & SW version */
      case CMD_GET_VER:
         commandBuffer[1] = VERSION_SW;
         commandBuffer[2] = VERSION_HW;
         commandBuffer[3] = ICP_Version_SW;
         commandBuffer[4] = ICP_Version_HW;
         return(5);     /* return cmd + 4 bytes of version */

      /* set target type */
      case CMD_SET_TARGET:
         rc = bdm_init(commandBuffer[2]);
         if (rc != BDM_RC_OK)
            commandBuffer[0] = rc;
         return (1);

      /* Set RS08 Derivative - horrible! */
      case CMD_SET_DERIVATIVE:
         bdm_option.derivative_type = commandBuffer[2];
         return(1);

      /* try to connect to the target */
      case CMD_CONNECT:
         rc = bdm_connect();
         if (rc != BDM_RC_OK)
            commandBuffer[0] = rc;
         return(1);

      /* request ICP on next reset */
      case CMD_SET_BOOT:  /* turn on bootloader & reset CPU! */
         if (strncmp((char*)&commandBuffer[2],"BOOT",sizeof("BOOT")-1) == 0) {
            forceICPReset();  /* Reboots to ICP mode - doesn't return */    
         }
         commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
         return(1);

      /* reset Target */
      case CMD_RESET:
         rc = bdm_targetReset(commandBuffer[2]);
         if (rc != BDM_RC_OK)
            commandBuffer[0] = rc;
         return(1);

      /* return status of BDM communication */
      case CMD_GET_STATUS:
         commandBuffer[1] = makeStatusByte();
         return(2);

      /* read BDM byte */
      case CMD_READ_BD: {
         U16 addr = ((commandBuffer[3]<<8)+commandBuffer[2]);
         if (addr != HC12_BDMSTS) {
            BDM12_CMD_BDREADB(addr,commandBuffer+1);
            return(2);
            }
         }
         // Fall through to CMD_READ_STATUS if reading BDM Status register
         goto cmd_read_status;
      
      case CMD_CF_READ_XCSR:
         // Fall through to CMD_READ_STATUS
         goto cmd_read_status;
#if 0
         BDMCF_CMD_READ_XCSR(commandBuffer+1);
         return(2);
#endif
      // Read BDM Status Register
      case CMD_READ_STATUS: {
      cmd_read_status:
         rc = 0;
         if (bdm_option.autoReconnect &&            // If auto re-connect enabled &
             ((bdm_status.target_type != HC12) ||   //   Target maybe supports SYNC
              (bdm_status.speed != SPEED_USER_SUPPLIED))) {
//              (bdm_status.speed == NO_INFO) ||
//              (bdm_status.speed == SYNC_SUPPORTED))) {
            rc = bdm_connect();                    // Make sure of connection
            }
         if (rc == BDM_RC_OK)
            rc = bdm_readBDMStatus(commandBuffer+1);
         if (rc != BDM_RC_OK) {
            commandBuffer[0] = rc;
            return 1;
         }
         return(2);
         }

      /* write BDM byte */
      case CMD_WRITE_BD: {
         U16 addr = ((commandBuffer[3]<<8)+commandBuffer[2]);
         if (addr != HC12_BDMSTS) {
            BDM12_CMD_BDWRITEB(addr,commandBuffer[4]);
            return(1);
            }
         }
      // Fall through to CMD_WRITE_STATUS if writing BDM Control register
      case CMD_CF_WRITE_XCSR:
#if 0
         BDMCF_CMD_WRITE_XCSR(commandBuffer[2]);
         return(1);
#endif
      // Write BDM Control Register
      case CMD_WRITE_CONTROL:
         rc = bdm_writeBDMControl(commandBuffer[2]);
         if (rc != BDM_RC_OK)
            commandBuffer[0] = rc;
         return(1);
      /* Stop the target */
      case CMD_HALT:
         bdm_halt();
         return(1);
      /* Mas erase RS08 Flash - not used? */
      case CMD_MASS_ERASE:
         rc = bdmRS08_massErase();
         if (rc != 0)
            commandBuffer[0] = rc;
         return(1) ;
      /* Set comm speed to user supplied value */
      case CMD_SET_SPEED1: {
         U16 sync_length = (U16)((commandBuffer[3]<<8)+commandBuffer[2]); // Save the new speed
         
         rc = bdmHC12_confirmSpeed(sync_length); // Confirm operation at that speed
         
         if (rc == 0) { // OK so far
             bdm_acknInit();               // Try ACKN feature
             rc = bdm_enableBDM();         // Try to enable BDM
         }
         if (rc != BDM_RC_OK) { // Failed
            commandBuffer[0] = rc;
            return(2);
         }
         bdm_status.speed = SPEED_USER_SUPPLIED; // User told us
         return(1);
      }
      /* Read current speed */
      case CMD_READ_SPEED1:
         commandBuffer[2] = (bdm_status.sync_length)>>8;   // Store speed in little endian
         commandBuffer[1] = (bdm_status.sync_length)&0xff;
         return(3);
      /* Start code execution */
      case CMD_GO1:
         //DEBUG_PIN = 1;
         bdm_go();
         return(1);

      /* Step over 1 instruction */
      case CMD_STEP1:
         bdm_step();
         return(1);

      /* read byte */
      case CMD_READ_8:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_READB((U16)256*commandBuffer[3]+commandBuffer[2],commandBuffer+1);
               return(2);
            case HCS08:
            case RS08:
               BDM08_CMD_READB((U16)((commandBuffer[3]<<8)+commandBuffer[2]), commandBuffer+1);
               return(2);
            }
         break;
      case CMD_READ_16:
         BDM12_CMD_READW((U16)256*commandBuffer[3]+commandBuffer[2],commandBuffer+2);
         commandBuffer[1] = commandBuffer[3];  /* swap bytes around for little endian */
         return(3);

      /* read registers */
      case CMD_READ_REGS:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_READ_PC(commandBuffer+2);
               commandBuffer[1] = commandBuffer[3];  // swap bytes around for little endian
               BDM12_CMD_READ_SP(commandBuffer+4);
               commandBuffer[3] = commandBuffer[5];  // swap bytes around for little endian
               BDM12_CMD_READ_X(commandBuffer+6);
               commandBuffer[5] = commandBuffer[7];  // swap bytes around for little endian
               BDM12_CMD_READ_Y(commandBuffer+8);
               commandBuffer[7] = commandBuffer[9];  // swap bytes around for little endian
               BDM12_CMD_READ_D(commandBuffer+10);
               commandBuffer[9] = commandBuffer[11]; // swap bytes around for little endian
               BDM12_CMD_BDREADW(HC12_BDMCCR,commandBuffer+11);  // CCR is in little endian format already
               commandBuffer[12] = 0;
               return(13);
            case HCS08:
               BDM08_CMD_READ_PC(commandBuffer+2);    
               commandBuffer[1] = commandBuffer[3];  // swap bytes around for little endian
               BDM08_CMD_READ_SP(commandBuffer+4);	 
               commandBuffer[3] = commandBuffer[5];  // swap bytes around for little endian
               BDM08_CMD_READ_HX(commandBuffer+6);    
               commandBuffer[5] = commandBuffer[7];  // swap bytes around for little endian
               BDM08_CMD_READ_A(commandBuffer+7);
               BDM08_CMD_READ_CCR(commandBuffer+8);					
               return(9);
            case RS08:
               BDM08_CMD_READ_PC(commandBuffer+2);    // RS08 Read CCR+PC
               commandBuffer[1] = commandBuffer[3];   // swap bytes around for little endian
               BDM08_CMD_READ_SP(commandBuffer+4);	   // RS08 Read Shadow PC
               commandBuffer[3] = commandBuffer[5];   // swap bytes around for little endian
               commandBuffer[5] = 0;                  // RS08 doesn't have Read HX
               commandBuffer[6] = 0;
               BDM08_CMD_READ_A(commandBuffer+7);     // RS08 Read A
               commandBuffer[8] = 0;                  // RS08 doesn't have Read CCR
               return(9);
            }
         break;
      // Read CFV1 General register
      case CMD_CF_READ_REG:
         BDMCF_CMD_READ_REGe(commandBuffer[2],commandBuffer+1);
         return(5);
      // Read CFV1 Debug register
      case CMD_CF_READ_DREG:
         BDMCF_CMD_READ_DREGe(commandBuffer[2],commandBuffer+1);
         return(5);
      // Read CFV1 Control register
      case CMD_CF_READ_CREG:
         BDMCF_CMD_READ_CREGe(commandBuffer[2],commandBuffer+1);
         return(5);
         
      // Write CFV1 General register
      case CMD_CF_WRITE_REG:
         BDMCF_CMD_WRITE_REGe(commandBuffer[2],(*(U32 *)(commandBuffer+3)));
         return(1);        
      // Write CFV1 Debug register
      case CMD_CF_WRITE_DREG:
         BDMCF_CMD_WRITE_DREGe(commandBuffer[2],(*(U32 *)(commandBuffer+3)));
         return(1);        
      // Write CFV1 Control register
      case CMD_CF_WRITE_CREG:
         BDMCF_CMD_WRITE_CREGe(commandBuffer[2],(*(U32 *)(commandBuffer+3)));
         return(1);        

      // Read CFV1 Memory
      case CMD_CF_READ_MEM: {
         U8 *data_ptr  = commandBuffer+1;    // Where in buffer to write the data
         U8 count      = commandBuffer[6];   // # of data bytes requested (excludes status byte)
         U8 orig_count = count;
         U8 memSpace   = commandBuffer[5];   // Size of the data writes
         U32u addr;                          // Address in target memory
         addr.bytes[0] = 0;
         addr.bytes[1] = commandBuffer[4];
         addr.bytes[2] = commandBuffer[3];
         addr.bytes[3] = commandBuffer[2];

#if (DEBUG&COMMAND_DEBUG)
         count--;    // Correct count for debug byte
#endif         
         if (count>MAX_PACKET_SIZE-1) {
            // requested block+status is too long to fit into the buffer
            commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS;
            return 1;
         }
         if (count >0) {
            switch (memSpace) {
               case 1:
                  BDMCF_CMD_READ_MEM_B(addr.longword,data_ptr);
                  while (count > 1) {
                     data_ptr       += 1;
                     addr.longword  += 1;
                     count          -= 1;
                     BDMCF_CMD_DUMP_MEM_B(data_ptr);
                  }
                  break;
               case 2:
                  BDMCF_CMD_READ_MEM_W(addr.longword,data_ptr);
                  while (count > 2) {
                     data_ptr       += 2;
                     addr.longword  += 2;
                     count          -= 2;
                     BDMCF_CMD_DUMP_MEM_W(data_ptr);
                  }
                  break;
               case 4:
                  BDMCF_CMD_READ_MEM_L(addr.longword,data_ptr);
                  while (count > 4) {
                     data_ptr       += 4;
                     addr.longword  += 4;
                     count          -= 4;
                     BDMCF_CMD_DUMP_MEM_L(data_ptr);
                  }
                  break;
               default:
                  commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS; 
                  break;
            }
         }
         return(orig_count+1);
      }

      // Write CFV1 Memory
      case CMD_CF_WRITE_MEM: {
         U8 *data_ptr  = commandBuffer+7;    // Where the data is
         U8 count      = commandBuffer[6];   // # of bytes
         U8 memSpace   = commandBuffer[5];   // Size of the data writes
         U32u addr;                          // Address in target memory
         addr.bytes[0] = 0;
         addr.bytes[1] = commandBuffer[4];
         addr.bytes[2] = commandBuffer[3];
         addr.bytes[3] = commandBuffer[2];
         if (count > 0) {
            switch (memSpace) {
               case 1:
                  BDMCF_CMD_WRITE_MEM_B(addr.longword,*data_ptr);
                  while (count > 1) {
                     data_ptr       += 1;
                     addr.longword  += 1;
                     count          -= 1;
                     BDMCF_CMD_FILL_MEM_B(*data_ptr);
                  }
                  break;
               case 2:
                  BDMCF_CMD_WRITE_MEM_W(addr.longword,*(U16 *)data_ptr);
                  while (count > 2) {
                     data_ptr       += 2;
                     addr.longword  += 2;
                     count          -= 2;
                     BDMCF_CMD_FILL_MEM_W(*(U16 *)data_ptr);
                  }
                  break;
               case 4:
                  BDMCF_CMD_WRITE_MEM_L(addr.longword,*(U32 *)data_ptr);
                  while (count > 4) {
                     data_ptr       += 4;
                     addr.longword  += 4;
                     count          -= 4;
                     BDMCF_CMD_FILL_MEM_L(*(U32 *)data_ptr);
                  }
                  break;
               default:
                  commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS; 
                  break;
            }
         }
         return(1);
      }

      /* read block of data from memory */
      case CMD_READ_BLOCK1: {
         // NOTE: This command returns block of BYTES, endianism does not matter here!!!
         U8  count      = commandBuffer[4];
         U8  count_orig = count;
         U16 addr       = (U16)((commandBuffer[3]<<8)+commandBuffer[2]);
         U8  *data_ptr  = commandBuffer+1;
         
         if (count>MAX_PACKET_SIZE-1) {
            // requested block+status is too long to fit into the buffer
            commandBuffer[0] = BDM_RC_ILLEGAL_PARAMS;
            return 1;
         }
         switch (bdm_status.target_type) {
            case HC12:
               while (count > 0) {
                  if ((addr&0x0001) || (count == 1)) {
                     // Address is odd or only 1 byte remaining
                     BDM12_CMD_READB(addr,data_ptr);  // fetch a byte
                     addr     +=1;                    // increment memory address
                     data_ptr +=1;                    // increment buffer pointer
                     count    -=1;                    // decrement count of bytes
                  }
                  else {
                     // Even address && >=2 bytes remaining
                     BDM12_CMD_READW(addr,data_ptr);  // fetch a word
                     addr     +=2;                    // increment memory address
                     data_ptr +=2;                    // increment buffer pointer
                     count    -=2;                    // decrement count of bytes
                  }
               }
               return(count_orig+1);
            case HCS08:
            case RS08:
               while(count>0) {
                  /* while there are bytes to fetch */
                  BDM08_CMD_READB(addr,data_ptr);  // fetch a byte
                  addr     +=1;                    // increment memory address
                  data_ptr +=1;                    // increment buffer pointer
                  count    -=1;                    // decrement count of bytes
               }
               return(count_orig+1);
            }
         break;
      }

      /* write block of bytes to memory */
      case CMD_WRITE_BLOCK1: {
         U8 count       = commandBuffer[4];
         U16 addr       = (U16)((commandBuffer[3]<<8)+commandBuffer[2]);
         U8 *data_ptr   = commandBuffer+5;

         switch (bdm_status.target_type) {
            case HC12:
               while (count > 0) {
                  if ((addr&0x0001) || (count == 1)) {
                     // Address is odd or only 1 byte remaining
                     BDM12_CMD_WRITEB(addr,*data_ptr);// write byte
                     addr     +=1;                    // increment memory address
                     data_ptr +=1;                    // increment buffer pointer
                     count    -=1;                    // decrement count of bytes
                  }
                  else {
                     // Even address && >=2 bytes remaining
                     BDM12_CMD_WRITEW(addr,*((U16 *)data_ptr)); // write a word
                     addr     +=2;                    // increment memory address
                     data_ptr +=2;                    // increment buffer pointer
                     count    -=2;                    // decrement count of bytes
                  }
               }
               return(1);
            case RS08:
#if (CAPABILITY & CAP_FLASH)
               if (flashState == FLASH_READY) {
                  //enableInterrupts(); // Allow USB operation while programming
                  (void)memcpy(flashProgrammingBuffer, commandBuffer, sizeof(flashProgrammingBuffer));
                  flashState = FLASH_BUSY; 
                  (void)bdmRS08_programFlash();
                  flashState = FLASH_READY;
                  disableInterrupts(); // Critical region
                  commandBuffer[0] = BDM_RC_OK;
                  return (1);
               }
               // Fall through otherwise
#endif // (CAPABILITY & CAP_FLASH)
            case HCS08:
               while(count>0) {
                  BDM08_CMD_WRITEB(addr,*data_ptr);
                  addr     +=1;                    // increment memory address
                  data_ptr +=1;                    // increment buffer pointer
                  count    -=1;                    // decrement count of bytes
               }
               return(1);
            }
         break;
      }

      /* write byte */
      case CMD_WRITE_8:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_WRITEB((U16)((commandBuffer[3]<<8)+commandBuffer[2]),commandBuffer[4]);
               return(1);
            case RS08:
               if (flashState == FLASH_READY) {
                  // Handle Flash programming of single bytes
                  flashProgrammingBuffer[2] = commandBuffer[2]; // Address to program
                  flashProgrammingBuffer[3] = commandBuffer[3];
                  flashProgrammingBuffer[4] = 1;                // # of bytes
                  flashProgrammingBuffer[5] = commandBuffer[4]; // Byte to program
                  flashState = FLASH_BUSY; 
                  (void)bdmRS08_programFlash();
                  flashState = FLASH_READY;
                  return (1);
               }
               // Fall through otherwise
            case HCS08:
               BDM08_CMD_WRITEB((U16)((commandBuffer[3]<<8)+commandBuffer[2]),commandBuffer[4]);
               return(1);
            }
         break;
      /* write word */
      case CMD_WRITE_16:
         BDM12_CMD_WRITEW((commandBuffer[3]<<8)+commandBuffer[2],(commandBuffer[5]<<8)+commandBuffer[4]);
         return(1);        
      /* write to PC */
      case CMD_WRITE_REG_PC:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_WRITE_PC((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case HCS08: // Write PC
            case RS08:  // Write CCR+PC
               BDM08_CMD_WRITE_PC((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            }
         break;
      /* write to SP */
      case CMD_WRITE_REG_SP:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_WRITE_SP((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case HCS08: // Write SP
            case RS08:  // Write shadow PC
               BDM08_CMD_WRITE_SP((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            }
         break;
      /* write to IX */
      case CMD_WRITE_REG_X:
         switch (bdm_status.target_type) {
            case HC12: // Write X
               BDM12_CMD_WRITE_X((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case HCS08: // Write HX
               BDM08_CMD_WRITE_HX((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case RS08: // Illegal - No Register H/X on RS08!
               commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
               return(1);
            }
         break;
      /* write to IY */
      case CMD_WRITE_REG_Y:
         switch (bdm_status.target_type) {
            case HC12: // Write Y
               BDM12_CMD_WRITE_Y((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case HCS08: // Illegal - No Register Y on HCS08!
            case RS08:  // Illegal - No Register Y on RS08!
               commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
               return(1);
            }
         break;
      /* write to B:A */
      case CMD_WRITE_REG_D:
         switch (bdm_status.target_type) {
            case HC12:
               BDM12_CMD_WRITE_D((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
               return(1);
            case HCS08:
            case RS08:
               BDM08_CMD_WRITE_A(commandBuffer[2]);
               return(1);
            }
         break;
      /* write to CCR */
      case CMD_WRITE_REG_CCR:
         /* CCR is stored in little endian notation */
         switch (bdm_status.target_type) {
            case HC12: // Write CCR
               BDM12_CMD_BDWRITEW(HC12_BDMCCR,*((U16 *)(commandBuffer+2))); 
               return(1);
            case HCS08:  // Write CCR
               BDM08_CMD_WRITE_CCR(commandBuffer[2]);
               return(1);
            case RS08:  // Illegal - Write CCR+PC is used instead
               commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
               return(2);
            }
         break;
      // Write to HCS08 Breakpoint reg
      case CMD_WRITE_BKPT:
         BDM08_CMD_WRITE_BKPT((U16)((commandBuffer[3]<<8)+commandBuffer[2]));
         return(1);
      // Read from HCS08 Breakpoint reg
      case CMD_READ_BKPT:
         BDM08_CMD_READ_BKPT(commandBuffer+2);	// Read BigEndian value
         commandBuffer[1] = commandBuffer[3];   // Convert to LittleEndian
         return(3);

#if (CAPABILITY & CAP_FLASH)
      // Prepare to write to RS08 Flash memory
      case CMD_FLASH_DLSTART:
         if (bdm_status.target_type != RS08) {
            commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
            return(1);
         }
         // Trim target clock to 8MHz
         rc = bdmRS08_trimTargetClock(SYNC_MULTIPLE(8000000UL));
         if (rc == 0)
            rc = bdm_connect();     // Re-connect after clock change
         if (rc != BDM_RC_OK) {
            commandBuffer[0] = rc;  // Failed trim or re-connect
            return 1;
         }
         bdm_halt();                // Stop processor
         rc = bdmRS08_vppOn();      // Turn VPP On
         if (rc != BDM_RC_OK) {
            commandBuffer[0] = rc;  // Failed to turn on Vpp
            return 1;
         }
         (void)bdmRS08_massErase(); // Mass erase the Flash
         flashState = FLASH_READY;  // We're ready to burn
         return(1);
      // Check if ready to write next block to RS08 Flash memory
      case TCMD_NEXT_DLDATA:
         if (bdm_status.target_type != RS08) {
            commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
            return(1);
         }
         if (flashState == FLASH_READY)
            commandBuffer[0] = BDM_RC_OK;
         else if (flashState == FLASH_BUSY)
            commandBuffer[0] = BDM_RC_FLASH_PROGRAMING_BUSY;
         else
            commandBuffer[0] = BDM_RC_WRONG_PROGRAMMING_MODE;
         return(1);
      // Finished programming RS08 Flash memory
      case CMD_FLASH_DLEND:
         if (bdm_status.target_type != RS08) {
            commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
            return(1);
         }
         if (flashState == FLASH_READY) { // Programming in progress?
            bdmRS08_restoreTrim();        // Re-program trim if saved
            VPP_OFF();                    // Turn Proramming voltage off
            flashState=FLASH_IDLE;        // Programming idle now
            (void)bdm_softwareReset(0);
            return(1);
         } 
         
         VPP_OFF();                  // Turn Programming voltage off
         flashState=FLASH_IDLE;      // Programming idle now
         
         commandBuffer[0] = BDM_RC_WRONG_PROGRAMMING_MODE;
         return (1);
#endif //(CAPABILITY & CAP_FLASH)
        
      /* unknown command */
      default: ;   
         // Do nothing - handled on exit from switch  
   }
   
   /* unknown or illegal command */
   commandBuffer[0] = BDM_RC_ILLEGAL_COMMAND;
   return(2);
}
#pragma MESSAGE DEFAULT C4000 // Restore warnings about always true
#pragma MESSAGE DEFAULT C5900 // Restore warnings about result is zero

//!  Processes all commands received over USB
//!
//!  @entry  The command is expected to be in commandBuffer[1]
//!  @return Number of bytes left in commandBuffer to be sent back as response
//!
U8 commandExec(void) {
#pragma DATA_SEG __SHORT_SEG Z_PAGE
static U8 busy = 0;
#pragma DATA_SEG DEFAULT
U8 size;

   if (busy)  {
      // Don't accept further commands until last one has completed
      // Problems here with async thread!!!!!
      commandBuffer[0] = BDM_RC_BUSY;
      return(1);
      }

   busy = 1;
   
   size = _commandExec();
   
   busy = 0;
   return size;
}

//!  Debug Layer
//!
//!  @entry The command is expected to be in commandBuffer[1]
//!  @return Number of bytes left in commandBuffer to be sent back as response
//!
U8 debugCommandExec(void){
U8 size;

   size = commandExec(); // Pass to actual execute routine

   // Append status byte to response
   commandBuffer[size++] = makeStatusByte();
   
   return size; 
}


//=========================================================================
//  Code below is not really used
//  Intended for asynchronous processing of commands - not yet implemented
//
void commandLoop(void) {
   deferredFunction = NULL;
   
   for(;;){
      wait();
      if (deferredFunction != NULL) {
         U8 size = deferredFunction();
         //disableInterrupts();
         
         }
   }
   
#if 0
   for(;;){
      int size;
      size = sizeof(commandBuffer);

      receiveCommand( &size, commandBuffer);
      size = commandExec();
      sendResponse( size, commandBuffer);
   }
#endif   
}