/*! \file
    \brief Simple USB Stack for JM60
    
   \verbatim
   UF32 USB Code
    
   Copyright (C) 2008  Peter O'Donoghue

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   \endverbatim

   \verbatim
Change Hisory
+====================================================================
| 24 Sep 2008 | fixed possible ptr error in ep0SaveOutData      - pgo
|  3 Mar 2008 | JM60 - USB code written from scratch            - pgo
+====================================================================
   \endverbatim
*/
#include <string.h>
#include <hidef.h>      /* for EnableInterrupts macro */
#include "derivative.h" /* include peripheral declarations */

#include "Common.h"
#include "Configure.h"
#include "BDM.h"
#include "BDMMacros.h"
#include "Commands.h"
#include "CmdProcessing.h"
#include "USBDefs.h"
#include "USB.h"

#if (DEBUG&COMMAND_DEBUG)
#define commandExec() debugCommandExec()
#endif

#define EP_OUT  (0x00)
#define EP_IN   (0x80)

//======================================================================
// Data packet odd/even indicator
enum {DATA0=0,DATA1=1};

//======================================================================
// Structure representing a BDT entry
typedef struct {
   union {
      struct {
         U8           :2;  
         U8 bdtkpid   :4;
         U8 data0_1   :1;
         U8 own       :1;
      } a;
      struct {
         U8           :2;  
         U8 bdtstall  :1;  
         U8 dts       :1;
         U8           :2;  
         U8 data0_1   :1;
         U8 own       :1;
      } b;
      U8 bits;
   } control;
   U8 byteCount;
   U8 epAddr;  // Offset by two bits 
} BDTEntry;

#define USB_INTMASKS (INTENB_STALL_MASK|INTENB_TOKDNE_MASK|INTENB_SOFTOK_MASK|INTENB_USBRST_MASK)

#define BDTEntry_OWN_MASK        (1<<7)
#define BDTEntry_DATA0_MASK      (0<<6) // dummy
#define BDTEntry_DATA1_MASK      (1<<6)
#define BDTEntry_DTS_MASK        (1<<3)
#define BDTEntry_BDTSTALL_MASK   (1<<2)

//======================================================================
// Maximum packet sizes for each endpoint
//
#define ENDPT0MAXSIZE    (64)    
#define ENDPT1MAXSIZE    (32)
#define ENDPT2MAXSIZE    (32)

//======================================================================
// Descriptors
//
static const DeviceDescriptor deviceDescriptor = {
   sizeof(DeviceDescriptor),               // Length (18 bytes)
   DT_DEVICE,                              // Type (device) (0x01)
   {0x10, 0x01},                           // USB spec rel. No. (BCD = 1.10)
   0xFF,                                   // Class code (none)
   0xFF,                                   // Sub Class code (none)
   0xFF,                                   // Protocol (none)
   ENDPT0MAXSIZE,                          // EndPt 0 max packet size
   {VendorID&0xFF,  (VendorID>>8)&0xFF},   // Vendor ID
   {ProductID&0xFF, (ProductID>>8)&0xFF},  // Product Id
   {0x00, 0x01},                           // Device Release (BCD = 1.00)
   1,                                      // String index of Manufacturer name
   2,                                      // String index of product desc.
   3,                                      // String index desc. serial #
   1                                       // Number of configurations
   };

static const struct {
   ConfigurationDescriptor configDescriptor;
   InterfaceDescriptor     interfaceDescriptor;
//   EndpointDescriptor      endpointDescriptor5;
//   EndpointDescriptor      endpointDescriptor6;
   } otherDescriptors = 
{
   { // configDescriptor
      sizeof(ConfigurationDescriptor),   // bLength = 9
      DT_CONFIGURATION,                  // bDescriptorType = 2
      {sizeof(otherDescriptors),0},      // wTotalLength = 25
      1,          // bNumInterfaces = 1
      1,          // bConfigurationValue
      0,          // iConfiguration = ""
      0x80,       // bmAttributes = Bus powered, no wakeup (yet?)
      150         // MaxPower = 300 mA
   },
   { // interfaceDescriptor
      sizeof(InterfaceDescriptor),  // bLength = 9
      DT_INTERFACE,                 // bDescriptorType = 4
      0,                            // bInterfaceNumber = 0
      0,                            // bAlternateSetting = 0
      0,                            // bNumEndpoints = 0
      0xFF,                         // bInterfaceClass = (Vendor specific)
      0xFF,                         // bInterfaceSubClass = (Vendor specific)
      0xFF,                         // bInterfaceProtocol = (Vendor specific)
      0                             // iInterface desc = ""
   },
#if 0
   { // endpointDescriptor1 - IN
     sizeof(EndpointDescriptor), // bLength = 7
     DT_ENDPOINT,                // bDescriptorType = 5
     EP_IN|1,                    // bEndpointAddress (IN+#1)
     0x02,                       // bmAttributes = Bulk
     {ENDPT1MAXSIZE,0},          // wMaxPacketSize
     1                           // bInterval = 1 ms
   },
   { // endpointDescriptor2 - OUT
     sizeof(EndpointDescriptor), // bLength = 7
     DT_ENDPOINT,                // bDescriptorType = 5
     EP_OUT|2,                   // bEndpointAddress (OUT+#2)
     0x02,                       // bmAttributes = Bulk
     {ENDPT2MAXSIZE,0},          // wMaxPacketSize = 16
     1                           // bInterval = 1 ms
   }
#endif
};

static const char sd0[] = {4,  DT_STRING, 0x08, 0x04}; // Language IDs
static const char sd1[] = {8,  DT_STRING, 'p', 0x00, 'g', 0x00, 'o', 0x00}; // Manufacturer
static const char sd2[] = {12, DT_STRING, 'U', 0x00, 'S', 0x00, 'B', 0x00, 'D', 0x00, 'M', 0x00}; // Description
static const char sd3[] = {12, DT_STRING, '0', 0x00, '0', 0x00, '0', 0x00, '0', 0x00, '1', 0x00}; // Serial #
static const char * const stringDescriptors[] = {sd0,sd1,sd2,(U8*)sd3};

//===============================================================================
// Device Status
//       
typedef struct {
   int selfPowered  : 1;
   int remoteWakeup : 1;
   int portTest     : 1;
   int res1         : 5;
   int res2         : 8;
} DeviceStatus; 

typedef enum { USBpowered, USBattached, USBdefault, USBaddressed,
               USBconfigured, USBsuspended } USBstateType;

#pragma DATA_SEG __SHORT_SEG Z_PAGE
struct {
   U8           state;
   U8           configuration;
   U8           interfaceAltSetting;
   DeviceStatus status;
   U8           newUSBAddress;
} deviceState = {USBattached, 0, 0, {0,0,0,0,0}, 0};
#pragma DATA_SEG DEFAULT

//===============================================================================
// Endpoint Status
//       
typedef struct {
   int stall  : 1;
   int res1   : 7;
   int res2   : 8;
} EPStatus;

// Endpoint state
typedef enum { 
   EPIdle = 0,       // Idle (Tx complete)
   EPDataIn,         // Doing a sequence of IN packets (until data count <= EPSIZE)
   EPDataOut,        // Doing a sequence of OUT packets (until data count <= EPSIZE)
   EPLastIn,         // Doing the last IN packet
   EPLastOut,        // Doing the last OUT packet
   EPStatusIn,       // Doing an IN packet as a status handshake
   EPStatusOut,      // Doing an OUT packet as a status handshake
   EPDataAvailable   // Data from last OUT is available, idle otherwise
} EPModes;

typedef struct {
   U8       state;
   U8       data0_1;
   EPStatus status;
   U8*      dataPtr;        // Pointer to data buffer 
   U8       dataRemaining;  // Count of remaining bytes to Rx/Tx
   U8       dataCount;      // Count of bytes Rx/Tx so far
   void     (*callback)(void);
} EPState;

#pragma DATA_SEG __SHORT_SEG Z_PAGE
static EPState ep0State = {EPIdle, DATA0, {0,0,0}, NULL, 0, 0, NULL};
//static EPState ep1State = {EPIdle, DATA0, {0,0,0}, NULL, 0, 0, NULL};
//static EPState ep2State = {EPIdle, DATA0, {0,0,0}, NULL, 0, 0, NULL};

static U8  usbActivityFlag = 0;
static U16 frameNum        = 0;

//======================================================================
// Buffer for EP0 Setup packet (copied from USB RAM)
SetupPacket ep0SetupBuffer;

#pragma DATA_SEG DEFAULT

//======================================================================
// USB RAM usage
//
#define USB_RAM_START (0x1860U)

#define USB_MAP_ADDRESS(x)  ((((x) - USB_RAM_START)>>2)&0xFC)
#define ADDRESS_ROUND16(x)  ((x+0xF)&0xFFF0)

//======================================================================
// BDTs for the endpoints (in USB RAM)
//
static BDTEntry bdts[10]    @USB_RAM_START;
#define  ep0BDTIn    bdts[0]
#define  ep0BDTOut   bdts[1]
#define  ep1BDT      bdts[2]
#define  ep2BDT      bdts[3]
#define  ep3BDT      bdts[4]
#define  ep4BDT      bdts[5]
#define  ep5BDTEven  bdts[6]
#define  ep5BDTOdd   bdts[7]
#define  ep6BDTEven  bdts[8]
#define  ep6BDTOdd   bdts[9]

//======================================================================
// Buffers for endpoint data packets (in USB RAM)
//
#define EP0InDataBufferAddress ADDRESS_ROUND16(USB_RAM_START+0x20)
static U8 ep0InDataBuffer[ENDPT0MAXSIZE]  @EP0InDataBufferAddress;
#define EP0OutDataBufferAddress ADDRESS_ROUND16(EP0InDataBufferAddress+ENDPT0MAXSIZE)
static U8 ep0OutDataBuffer[ENDPT0MAXSIZE] @EP0OutDataBufferAddress;
//#define EP1DataBufferAddress ADDRESS_ROUND16(EP0OutDataBufferAddress+ENDPT0MAXSIZE)
//static U8 ep1DataBuffer[ENDPT1MAXSIZE]    @EP1DataBufferAddress;
//#define EP2DataBufferAddress ADDRESS_ROUND16(EP1DataBufferAddress+ENDPT1MAXSIZE)
//static U8 ep2DataBuffer[ENDPT2MAXSIZE]    @EP2DataBufferAddress;

//======================================================================
// All of the USB RAM
//
static volatile U8 usbRam[256] @USB_RAM_START;

#pragma DATA_SEG DEFAULT
//======================================================================
// Configure the BDT for EP0 Out (Rx, device <- host, DATA0/1)
//
void ep0InitialiseBDTOut(void){
U16 size;

   // Calculate max size of next DATA0/1 pkt
   size = ep0State.dataRemaining; 
   if (size > ENDPT0MAXSIZE)
      size = ENDPT0MAXSIZE;
   
   // Set up to Rx packet
   ep0BDTOut.epAddr      = USB_MAP_ADDRESS(EP0OutDataBufferAddress);
   ep0BDTOut.byteCount   = (U8)size;
   if (ep0State.data0_1) 
      ep0BDTOut.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA1_MASK|BDTEntry_DTS_MASK;
   else
      ep0BDTOut.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA0_MASK|BDTEntry_DTS_MASK;
}

//=========================================================================
// Save the data from an EP0 OUT pkt and advance ptrs etc.
//
void ep0SaveOutData(U8 *dataPtr) {
   // Copy the date from the Rx buffer
   if ((ep0BDTOut.byteCount>0) && (dataPtr != NULL))
      (void)memcpy(dataPtr, ep0OutDataBuffer, ep0BDTOut.byteCount);
   ep0State.dataRemaining -= ep0BDTOut.byteCount;         // Count down bytes to go
   ep0State.dataCount     += ep0BDTOut.byteCount;         // Count bytes so far
   ep0State.dataPtr        = dataPtr+ep0BDTOut.byteCount;
   ep0State.data0_1        = !ep0State.data0_1;           // Toggle DATA0/1
}

//======================================================================
// Configure EP0 for an OUT transaction (Rx, device <- host, DATA0/1)
//
void ep0StartOutTransaction( U8 bufSize, U8 *bufPtr, U8 data0_1 ){

   ep0State.dataRemaining = bufSize; // Total bytes to Rx
   ep0State.dataCount     = 0;       // Reset count of bytes so far
   ep0State.dataPtr       = bufPtr;  // Where to (eventually) place data
   ep0State.data0_1       = data0_1; // Initial data toggle
   
   if (bufSize == 0) 
      ep0State.state = EPStatusOut;  // Assume status handshake
   else if (bufSize < ENDPT0MAXSIZE) 
      ep0State.state = EPLastOut;    // Assume one and only data pkt
   else
      ep0State.state = EPDataOut;    // Assume first of several data pkts

   ep0InitialiseBDTOut(); // Configure the BDT for transfer
}

//======================================================================
// Configure EP0 for a SETUP transaction (Rx, device<-host, DATA0)
//
void ep0StartSetupTransaction( void ){

   ep0StartOutTransaction(sizeof(ep0SetupBuffer), (U8*)&ep0SetupBuffer,  DATA0);
   ep0State.state = EPIdle;         // Now Idle
}

//======================================================================
// Configure the BDT for EP0 In (Tx, device -> host)
//
void ep0InitialiseBDTIn( void ){
U16 size;

   // Enable endpoint
   EPCTL0  = EPCTL0_EPRXEN_MASK|EPCTL0_EPTXEN_MASK|EPCTL0_EPHSHK_MASK;

   size = ep0State.dataRemaining;
   if (size > ENDPT0MAXSIZE)
      size = ENDPT0MAXSIZE;
   
   // Copy the date to Tx buffer
   (void)memcpy(ep0InDataBuffer, ep0State.dataPtr, size);
      
   ep0State.dataPtr         += size;  // Ptr to _next_ data
   ep0State.dataRemaining   -= size;  // Count of remaining bytes
   ep0State.dataCount       += size;  // Count of bytes so far
   
   // Set up to Tx packet
   ep0BDTIn.epAddr        = USB_MAP_ADDRESS(EP0InDataBufferAddress);
   ep0BDTIn.byteCount     = (U8)size;
   if (ep0State.data0_1) 
      ep0BDTIn.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA1_MASK|BDTEntry_DTS_MASK;
   else
      ep0BDTIn.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA0_MASK|BDTEntry_DTS_MASK;

   ep0State.data0_1         = !ep0State.data0_1;   // Toggle DATA0/1 for next pkt
}

//======================================================================
// Configure EP0 for an IN transaction (Tx, device -> host, DATA0/1)
//
void ep0StartInTransaction( U8 bufSize, const U8 *bufPtr, U8 data0_1 ){

   ep0State.dataPtr         = (U8*)bufPtr;   // Ptr to _next_ data
   ep0State.dataRemaining   = bufSize;       // Count of remaining bytes
   ep0State.dataCount       = 0;             // Reset count of bytes so far
   ep0State.data0_1         = data0_1;       // Initial data toggle
   
   if (bufSize == 0) 
      ep0State.state = EPStatusIn;   // Assume status handshake
   else if (bufSize < ENDPT0MAXSIZE) 
      ep0State.state = EPLastIn;     // Assume one and only data pkt
   else
      ep0State.state = EPDataIn;     // Assume first of several data pkt
      
   ep0InitialiseBDTIn(); // Configure the BDT for transfer
}
#ifdef USE_USBPT1
//======================================================================
// Configure the BDT for EP1 In
//
void ep1InitialiseBDTIn( void ){
U16 size;

   size = ep1State.dataRemaining;
   if (size > ENDPT1MAXSIZE)
      size = ENDPT1MAXSIZE;
   
   // Copy the date to Tx buffer
   (void)memcpy(ep1DataBuffer, ep1State.dataPtr, size);
      
   ep1State.dataPtr         += size;  // Ptr to _next_ data
   ep1State.dataRemaining   -= size;  // Count of remaining bytes
   ep1State.dataCount       += size;  // Count of bytes so far

   // Set up to Tx packet
   ep1BDT.epAddr          = USB_MAP_ADDRESS(EP1DataBufferAddress);
   ep1BDT.byteCount       = (U8) size;
   if (ep1State.data0_1) 
      ep1BDT.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA1_MASK|BDTEntry_DTS_MASK;
   else
      ep1BDT.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA0_MASK|BDTEntry_DTS_MASK;

   ep1State.data0_1         = !ep1State.data0_1;   // Toggle DATA0/1 for next pkt
}

//======================================================================
// Configure EP1 for an IN transaction (Tx, device -> host, DATA0/1)
//
void ep1StartInTransaction( U16 bufSize, const U8 *bufPtr ){

   ep1State.dataPtr       = (U8 *)bufPtr;
   ep1State.dataRemaining = bufSize;
   ep1State.dataCount     = 0;                   // Reset count of bytes so far
   ep1State.state         = EPDataIn;
   
   ep1InitialiseBDTIn();
}

//======================================================================
// Configure the BDT for EP2 Out
//
void ep2InitialiseBDTOut( void ){

U16 size;

   size = ep2State.dataRemaining;
   if (size > ENDPT2MAXSIZE)
      size = ENDPT2MAXSIZE;
   
   ep2State.state          = EPDataOut;

   // Set up to Tx packet
   ep2BDT.epAddr           = USB_MAP_ADDRESS(EP2DataBufferAddress);
   ep2BDT.byteCount        = (U8)size;
   if (ep1State.data0_1) 
      ep2BDT.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA1_MASK|BDTEntry_DTS_MASK;
   else
      ep2BDT.control.bits  = BDTEntry_OWN_MASK|BDTEntry_DATA0_MASK|BDTEntry_DTS_MASK;
}

void ep2SaveOutData(void) {
   // Copy the date from the Rx buffer
   (void)memcpy(ep2State.dataPtr, ep2DataBuffer, ep2BDT.byteCount);
   ep2State.dataRemaining -= ep2BDT.byteCount;    // Count bytes Received
   ep2State.dataCount     += ep2BDT.byteCount;
   ep2State.dataPtr       += ep2BDT.byteCount;    // Advance data ptr
   ep2State.data0_1        = !ep2State.data0_1;   // toggle DATA0/1 for next pkt
}

//======================================================================
// Configure EP2 for an OUT transaction (Rx, device <- host, DATA0/1)
//
void ep2StartOutTransaction( U16 bufSize, U8 *bufPtr ){

   ep2State.dataPtr        = bufPtr;
   ep2State.dataRemaining  = bufSize;
   ep2State.dataCount      = 0;       // Reset count of bytes so far

   ep2InitialiseBDTOut();
}
#endif

//======================================================================
// Stall control for endpoints
//

// Stall ep0
void ep0Stall(void) {
   EPCTL0 |= EPCTL0_EPSTALL_MASK; 
   ep0State.status.stall = 1;
}

#ifdef USE_USBPT1
// Stall ep1
void ep1Stall(void) {
   EPCTL1 |= EPCTL1_EPSTALL_MASK; 
   ep1State.status.stall = 1;
}

// Stall ep2
void ep2Stall(void) {
   EPCTL2 |= EPCTL2_EPSTALL_MASK; 
   ep2State.status.stall = 1;
}
#endif

// Clear stall ep0
void ep0ClearStall(void) {
   EPCTL0 &= ~EPCTL0_EPSTALL_MASK; 
   ep0State.status.stall = 0;
}

#ifdef USE_USBPT1
// Clear stall ep1
void ep1ClearStall(void) {
   EPCTL1 &= ~EPCTL1_EPSTALL_MASK; 
   ep1State.status.stall = 0;
}

// Clear stall ep2
void ep2ClearStall(void) {
   EPCTL2 &= ~EPCTL2_EPSTALL_MASK; 
   ep2State.status.stall = 0;
}

// Clear Data toggle (set to DATA0 on next Tx/Rx)
#pragma INLINE
void ep1ClearDataToggle(){
   ep1State.data0_1 = DATA0;
}

// Clear Data toggle (set to DATA0 on next Tx/Rx)
#pragma INLINE
void ep2ClearDataToggle(){
   ep2State.data0_1 = DATA0;
}
#endif

//========================================================================================
//
void setUSBdefaultState( void ){
   GREEN_LED_OFF();
   deviceState.state                = USBdefault;
   ADDR                             = 0;
   deviceState.configuration        = 0;
   deviceState.interfaceAltSetting  = 0; 
}

void setUSBaddressedState( U8 address ){
   if (address == 0) // Unaddress??
      setUSBdefaultState();
   else {
      GREEN_LED_OFF();
      deviceState.state                = USBaddressed;
      ADDR                             = address;
      deviceState.configuration        = 0;
      deviceState.interfaceAltSetting  = 0; 
   }
}

void setUSBconfiguredState( U8 config ){
   if (config == 0) // unconfigure
      setUSBaddressedState(ADDR);
   else {
      GREEN_LED_ON();
      deviceState.state                = USBconfigured;
      deviceState.configuration        = config;
      deviceState.interfaceAltSetting  = 0; 
#ifdef USE_USBPT1
      ep1ClearDataToggle();
      ep2ClearDataToggle();
      ep1ClearStall();
      ep2ClearStall();
#endif
   }
}

//======================================================================
// Initialise the USB interface
//
void initUSB( void ){

   // Clear USB RAM (includes BDTs)
   (void)memset(usbRam, 0x00, sizeof(usbRam));
   
   // Reset USB   
   USBCTL0_USBRESET = 1;
   while (USBCTL0_USBRESET) {
   }
   
   // Enable USB module.
   CTL = CTL_USBEN_MASK;
   
   // Internal PUP, Internal 3V3 reg, Enable Txvr 
   USBCTL0 = USBCTL0_USBPU_MASK|USBCTL0_USBVREN_MASK|USBCTL0_USBPHYEN_MASK;

   // Clear all pending interrupts except reset.
   INTSTAT = ~INTSTAT_USBRSTF_MASK;
   
   // Disable error interrupts.   
   ERRENB=0;

   // Enable various interrupts
   INTENB  = USB_INTMASKS;
   
   deviceState.state = USBdefault;
   ep0State.state = EPIdle;
//   ep1State.state = EPIdle;
//   ep2State.state = EPIdle;

   ep0ClearStall();

   // Set up to receive setup packet
   ep0StartSetupTransaction();
 
   // Enable endpoints
   EPCTL0  = EPCTL0_EPRXEN_MASK|EPCTL0_EPTXEN_MASK|EPCTL0_EPHSHK_MASK;
//   EPCTL1  = EPCTL1_EPTXEN_MASK|EPCTL1_EPHSHK_MASK;
//   EPCTL2  = EPCTL2_EPRXEN_MASK|EPCTL2_EPHSHK_MASK;

   // Enable usb reset interrupt
   //INTENB_USBRST=1;
  
   setUSBdefaultState();
   }

//===============================================================================
// Get Status - Device Req 0x00
//       
void handleGetStatus( void ) {
static const U8 zeroReturn[] = {0,0};

   switch(ep0SetupBuffer.bmRequestType) {
      case (EP_IN|RT_DEVICE) : // Device Status
         ep0StartInTransaction( sizeof(deviceState.status), (char *) &deviceState.status, DATA1 );
         break;
      case (EP_IN|RT_INTERFACE) : // Interface Status - reserved
         ep0StartInTransaction( sizeof(zeroReturn), (char *) &zeroReturn, DATA1 );
         break;
      case (EP_IN|RT_ENDPOINT) : // Endpoint Status
#if 0
         if (index == (EP_IN|1)) {
            ep0StartInTransaction( sizeof(ep1State.status), (char *) &ep1State.status, DATA1 );
            }
         else if (index == 2) {
            ep0StartInTransaction( sizeof(ep2State.status), (char *) &ep2State.status, DATA1 );
            }
         else 
#endif
         {
            ep0Stall();
            break;
            }
         break;
      default : // Illegal
         ep0Stall();
         break;
      }
}    

//===============================================================================
// Clear Feature - Device Req 0x01
//       
void handleClearFeature( void ) {
U16 featureSelector  = (ep0SetupBuffer.wValue.le.hi<<8) | ep0SetupBuffer.wValue.le.lo;
int index            = (ep0SetupBuffer.wIndex.le.hi<<8) | ep0SetupBuffer.wIndex.le.lo;
int okStatus         = 0;

   switch(ep0SetupBuffer.bmRequestType) {
      case (RT_DEVICE) : // Device Feature
         if ((featureSelector != DEVICE_REMOTE_WAKEUP) || // Device remote wakeup
             (index != 0))   {                            // device index must be 0
            break;
            }
         deviceState.status.remoteWakeup = 0;
         okStatus                        = 1;
         break;

      case (RT_INTERFACE) : // Interface Feature
         break;

      case (RT_ENDPOINT) : // Endpoint Feature ( Out,Standard,Endpoint )
         if (featureSelector != ENDPOINT_HALT) // Not Endpoint Stall ?
            break;
#ifdef USE_USBPT1
         if (index == (EP_IN|1)) {   // endpt 1
            ep1ClearDataToggle();
            ep1ClearStall(); // clear stall on Endpt 1
            okStatus = 1;
            }
         else if (index == (EP_OUT|2)) { // endpt 2
            ep2ClearDataToggle();
            ep2ClearStall();    // clear stall on Endpt 2
            okStatus = 1;
            }
#endif
         break;

      default : // Illegal
         break;
      }

   if (okStatus) {
      ep0StartInTransaction( 0, NULL, DATA1 ); // Tx empty Status packet
      }
   else
      ep0Stall();
}

//===============================================================================
// Set Feature - Device Req 0x03
//       
void handleSetFeature( void ) {
U16 featureSelector = (ep0SetupBuffer.wValue.le.hi<<8) | ep0SetupBuffer.wValue.le.lo;
int index           = (ep0SetupBuffer.wIndex.le.hi<<8) | ep0SetupBuffer.wIndex.le.lo;
int okStatus        = 0;

   switch(ep0SetupBuffer.bmRequestType) {
      case (RT_DEVICE) : // Device Feature
         if ((featureSelector != DEVICE_REMOTE_WAKEUP) || // Device remote wakeup
             (index != 0)) {                              // device index must be 0
            break;
            }
         deviceState.status.remoteWakeup = 1;
         okStatus                        = 1;
         break;

      case (RT_INTERFACE) : // Interface Feature
         break;

      case (RT_ENDPOINT) : // Endpoint Feature ( Out,Standard,Endpoint )
#ifdef USE_USBPT1
         if (featureSelector != ENDPOINT_HALT) // Not Endpoint Stall ?
            break;
         if (index == (EP_IN|1)) { // endpt 1
            ep1Stall();
            okStatus              = 1;
            }
         else if (index == (EP_OUT|2)) { // endpt 2
            ep2Stall();
            okStatus              = 1;
            }
#endif
         break;

      default : // Illegal
         break;
      }
      
   if (okStatus) {
      ep0StartInTransaction( 0, NULL, DATA1 ); // Tx empty Status packet
      }
   else
      ep0Stall();
}

//===============================================================================
// Get Descriptor - Device Req 0x06
//       
void handleGetDescriptor( void ) {
int         Slength         = (ep0SetupBuffer.wLength.le.hi<<8) | ep0SetupBuffer.wLength.le.lo;
int         descriptorIndex = ep0SetupBuffer.wValue.le.lo;
int         dataSize = 0;
const U8   *dataPtr = NULL;

   if (ep0SetupBuffer.bmRequestType != (EP_IN|RT_DEVICE)) {// In,Standard,Device
      ep0Stall();
      return;
      }

   switch (ep0SetupBuffer.wValue.le.hi) {

      case DT_DEVICE: // Get Device Desc. - 1
         dataSize = deviceDescriptor.bLength;
         dataPtr  = (U8 *) &deviceDescriptor;
         break;
      case DT_CONFIGURATION: // Get Configuration Desc. - 2
         dataSize = otherDescriptors.configDescriptor.wTotalLength.le.lo;
         dataPtr  = (U8 *) &otherDescriptors;
         break;
      case DT_STRING: // Get String Desc.- 3
         if (descriptorIndex >= sizeof(stringDescriptors)/sizeof(stringDescriptors[0])) {
            ep0Stall(); // Illegal string index - stall
            return;
            }
         dataPtr  = (U8*) stringDescriptors[descriptorIndex];
         dataSize = *dataPtr;
         break;
      default: // shouldn't happen
         ep0Stall();
         return;
      } // switch

   if (dataSize > Slength) // Truncate if more bytes than asked for
      dataSize = Slength;
      
   ep0StartInTransaction( (U8)dataSize, dataPtr, DATA1 ); // Set up Tx
}

//===============================================================================
// Set device Address Callback
//       
static void setAddressCallback(void) {
   ADDR = deviceState.newUSBAddress;
   if ((deviceState.state == USBdefault) && (deviceState.newUSBAddress != 0))
      setUSBaddressedState(deviceState.newUSBAddress);
   else if ((deviceState.state == USBaddressed) && (deviceState.newUSBAddress == 0))
      setUSBdefaultState();
}

//===============================================================================
// Set device Address - Device Req 0x05
//       
void handleSetAddress( void ) {

   if (ep0SetupBuffer.bmRequestType != (0x00|RT_DEVICE)) {// In,Standard,Device
      ep0Stall(); // Illegal format - stall ep0
      return;
      }
   deviceState.newUSBAddress  = ep0SetupBuffer.wValue.le.lo; // Save address for change after status transaction
   ep0State.callback          = setAddressCallback;
   
   ep0StartInTransaction( 0, NULL, DATA1 ); // Tx empty Status packet
}

//===============================================================================
// Get Configuration - Device Req 0x08
//       
void   handleGetConfiguration( void ) {
//static U8 configBuffer;

//   configBuffer = (U8)deviceState.configuration;  // set up buffer for transfer
   ep0StartInTransaction( 1, (U8 *) &deviceState.configuration, DATA1 );
}

//===============================================================================
// Set Configuration - Device Req 0x09
//       
void handleSetConfiguration( void ) {

   if ((ep0SetupBuffer.bmRequestType != (0x00|RT_DEVICE)) || // Out,Standard,Device
       ((ep0SetupBuffer.wValue.le.lo != 0) &&       // Only support 1 configuration
        (ep0SetupBuffer.wValue.le.lo != otherDescriptors.configDescriptor.bConfigurationValue))) { 
      ep0Stall();
      return;
      }
   
   setUSBconfiguredState(ep0SetupBuffer.wValue.le.lo);
   
   ep0StartInTransaction( 0, NULL, DATA1 ); // Tx empty Status packet
}

//===============================================================================
// Get interface - Device Req 0x0A
//       
void handleGetInterface( void ) {
static U8 interfaceBuffer;
int Slength        = (ep0SetupBuffer.wLength.le.hi<<8) + ep0SetupBuffer.wLength.le.lo;
int interfaceIndex = (ep0SetupBuffer.wValue.le.hi<<8)  + ep0SetupBuffer.wValue.le.lo;
U8 reqType       = ep0SetupBuffer.bmRequestType;

   if ((ep0SetupBuffer.bmRequestType != (EP_IN|RT_INTERFACE)) || // In,Standard,Interface
       (Slength != 1)) {
      ep0Stall();
      return;
      }

   // only support one interface
   if (interfaceIndex != otherDescriptors.interfaceDescriptor.bInterfaceNumber) {
      ep0Stall();
      return;
      }

   interfaceBuffer = (U8)deviceState.interfaceAltSetting;  // set up buffer for transfer
   ep0StartInTransaction( 1, (char *) &interfaceBuffer, DATA1 ); // Send packet
}

//=================================================
// Illegal request in SETUP pkt
//
void handleUnexpected(void) {
   ep0Stall();
}

//===============================================================================
// Process BDM command Callback
//       
void processCommandCallback(void) {
   (void)commandExec(); // Execute the command
}

//===============================================================================
// Handles SETUP Packet
//       
void handleSetupToken( void ) {
static void (* const(funcs[]))(void) = {
   handleGetStatus,        //   0 : GET_STATUS 
   handleClearFeature,     //   1 : CLEAR_FEATURE
   handleUnexpected,       //   2 : Reserved
   handleSetFeature,       //   3 : SET_FEATURE
   handleUnexpected,       //   4 : Reserved
   handleSetAddress,       //   5 : SET_ADDRESS
   handleGetDescriptor,    //   6 : GET_DESCRIPTOR
   handleUnexpected,       //   7 : SET_DESCRIPTOR - not supported
   handleGetConfiguration, //   8 : GET_CONFIGURATION
   handleSetConfiguration, //   9 : SET_CONFIGURATION
   handleGetInterface,     //  10 : GET_INTERFACE
   handleUnexpected,       //  11 : SET_INTERFACE - not supported
   handleUnexpected,       //  12 : SYNCH_FRAME
   };
U8 size;
U8 dataSize;
   
   ep0SaveOutData((U8*)&ep0SetupBuffer); // Save setup pkt

   ep0StartSetupTransaction();   // In case another SETUP pkt
   ep0State.callback = NULL;

   if (IS_VENDOR_REQ(ep0SetupBuffer.bmRequestType)) {
      commandBuffer[1] = ((U8*)&ep0SetupBuffer)[1];
      commandBuffer[2] = ((U8*)&ep0SetupBuffer)[2];
      commandBuffer[3] = ((U8*)&ep0SetupBuffer)[3];
      commandBuffer[4] = ((U8*)&ep0SetupBuffer)[4];
      commandBuffer[5] = ((U8*)&ep0SetupBuffer)[5];
      // CMD_CF_READ_MEM doesn't have room for the block size in the SETUP pkt so
      // its calculated here! Placement is consistent with CMD_CF_WRITE_MEM
      commandBuffer[6] = ((U8*)&ep0SetupBuffer)[6]-1; // data size = wlength.le.lo-1
      commandBuffer[7] = 0;                           // to be consistent with other commands
      dataSize = ep0SetupBuffer.wLength.le.lo;
      if (ep0SetupBuffer.wLength.le.hi != 0)
         dataSize = 255;
      if (ep0SetupBuffer.bmRequestType&EP_IN) { // Command + Data In
         size = commandExec();
         if (size > dataSize) // Make sure we don't return too much data
            size = (U8)dataSize;
         // Set up to send data
         ep0StartInTransaction( size, commandBuffer, DATA1 );
      }
      else { // Command + Data Out
         if (dataSize == 0) { // No Data phase?
            size = commandExec();
            ep0StartInTransaction( 0, NULL, DATA1 ); // Tx empty Status packet
            }
         else {
            // Set up to get rest of command (data portion)
            ep0StartOutTransaction( dataSize, commandBuffer+6, DATA1 );
            // Set callback to execute when data arrives
            ep0State.callback = processCommandCallback;
         }
      }
   }   
   else if ( ep0SetupBuffer.bRequest <= sizeof(funcs)/sizeof(funcs[0])) { // Valid request?
      funcs[ep0SetupBuffer.bRequest]();     // Dispatch to service routine
   }
   else
      handleUnexpected();
      
   // Allow transactions post SETUP
   CTL_TSUSPEND = 0;
}

//==================================================================
// Handlers for Token Complete USB interrupt
//

//================
// ep0 - IN
// 
void handleEp0InToken(void){

   switch (ep0State.state) {
      case EPDataIn:    // Doing a sequence of IN packets (until data count <= EPSIZE)
         if (ep0State.dataRemaining < ENDPT0MAXSIZE) // Last pkt?
            ep0State.state = EPLastIn;    // Sending an empty/undersize pkt
         else 
            ep0State.state = EPDataIn;    // Sending full pkt
         ep0InitialiseBDTIn(); // Set up next IN pkt
         break;
         
      case EPLastIn:    // Just done the last IN packet
         ep0State.state = EPStatusOut;   // Receiving an OUT status pkt
         ep0StartOutTransaction(0, NULL, DATA1); // Do status Pkt reception
         break;
         
      case EPStatusIn:  // Just done an IN packet as a status handshake for an OUT Data transfer
         if (ep0State.callback != NULL)
            ep0State.callback(); // Execute callback function to process OUT data
         ep0State.callback = NULL;
         ep0State.state    = EPIdle;
         break;

      // We don't expect an IN token while in the following states
      case EPIdle:           // Idle (Tx complete)
      case EPDataOut:        // Doing a sequence of OUT packets (until data count <= EPSIZE)
      case EPLastOut:        // Doing the last OUT packet
      case EPStatusOut:      // Doing an OUT packet as a status handshake
      case EPDataAvailable:  // Data from last OUT is available, idle otherwise
         break;
   }
}

//================
// ep0 - OUT
// 
void handleEp0OutToken(void){

   switch (ep0State.state) {
      case EPDataOut:        // Doing a sequence of OUT packets (until data count <= EPSIZE)
         ep0SaveOutData(ep0State.dataPtr); // Save the data from the Rx buffer
         if (ep0State.dataRemaining < ENDPT0MAXSIZE) // Last pkt?
            ep0State.state = EPLastOut;    // Receiving an empty/undersize pkt
         else 
            ep0State.state = EPDataOut;    // Receiving full pkt
         ep0InitialiseBDTOut(); // Set up next OUT pkt
         break;

      case EPLastOut:        // Done the last OUT packet
         ep0SaveOutData(ep0State.dataPtr);      // Save the data from the Rx buffer
         ep0State.state = EPDataAvailable;      // Receiving full pkt
         ep0StartSetupTransaction();            // Make ready for a SETUP pkt
         ep0StartInTransaction(0, NULL, DATA1); // Do status Pkt transmission
         break;

      case EPStatusOut:      // Doing an OUT packet as a status handshake
         ep0StartSetupTransaction();      // Make ready for a SETUP pkt
         break;
        
      // We don't expect an OUT token while in the following states
      case EPLastIn:          // Just done the last IN packet
      case EPDataIn:          // Doing a sequence of IN packets (until data count <= EPSIZE)
      case EPStatusIn:        // Just done an IN packet as a status handshake
      case EPIdle:            // Idle (Tx complete)
      case EPDataAvailable:   // Data from last OUT is available, idle otherwise
         ep0StartSetupTransaction();      // Discard & make ready for a SETUP pkt
         break;
   }
}

//=================================================
// ep0 - STALL completed - re-enable ep0 for SETUP
// 
void handleEp0StallComplete(void){
   if (EPCTL0 & EPCTL0_EPSTALL_MASK) // ep0 stalled? 
      ep0StartSetupTransaction(); // yes - re-enable
}

#ifdef USE_USBPT1
//================
// ep1 - IN
// 
void handleEp1InToken(void){
   if (ep1State.state == EPDataIn) {
      if (ep1State.dataRemaining==0) // Finished?
         ep1State.state = EPIdle; // now idle
      else
         ep1InitialiseBDTIn();
   }
}

//================
// ep2 - OUT
// 
void handleEp2OutToken(void){
   if (ep2State.state == EPDataOut) { 
      // Save the data from the buffer
      ep2SaveOutData();
      
//      if (ep2State.dataRemaining==0) { // Finished?
      if (ep2BDT.byteCount<ENDPT2MAXSIZE) // Short Pkt => Finished?
         ep2State.state = EPDataAvailable;  // Now idle with data available
      else
         ep2InitialiseBDTOut();
   }
}
#endif

//==================================================================
// Handler for Token Complete USB interrupt
//
// Only handles ep0 (In & Out), ep1 (In) & ep2 (Out)
// 
void handleTokenComplete(U8 status){
   
   U8 endPoint      = ((U8)status)>>4;             // Endpoint number
   U8 directionIsIn = status&STAT_IN_MASK;         // Direction of T/F 0=>OUT, (!=0)=>IN
   U8 receivedPID   = ep0BDTOut.control.a.bdtkpid; // PID 

   if (endPoint == 0) { // Accept IN, OUT or SETUP token
      if (receivedPID == SETUPToken) // SETUP transaction complete
         handleSetupToken();
      else if (directionIsIn)  // IN Transaction complete
         handleEp0InToken();
      else // OUT Transaction
         handleEp0OutToken();
      return;
   }

#ifdef USE_USBPT1
   if (deviceState.state < USBconfigured) { // We can't accept any tokens
      // Stall the endpoint associated with the unexpected token??
      if (endPoint == 1)
         ep1Stall();
      if (endPoint == 2)
         ep2Stall();
      return;
   }
   
   if (endPoint == 1) {
      if (directionIsIn)  // IN Transaction complete
         handleEp1InToken();
      else
         ep1Stall();
      return;
   }

   if (endPoint == 2) {
      if (!directionIsIn)  // OUT Transaction complete
         handleEp2OutToken();
      else
         ep2Stall();
      return;
   }
#endif
}

#pragma MESSAGE DISABLE C4003
//==================================================================
// Handler for Start of Frame Token interrupt
//
void handleSOFToken() {
static U8 flashCount = 0;

   asm {
      lda FRMNUMH
      sta frameNum
      lda FRMNUML
      sta frameNum:1
   }
   // Green LED
   // Off                     - no USB activity, not connected
   // On                      - no USB activity, connected
   // Off, flash briefly on   - USB activity, not connected
   // On,  flash briefly off  - USB activity, connected
   if (FRMNUML==0) { // Every ~256 ms
      switch (flashCount++) {
         case 0:
            if (deviceState.state == USBconfigured)
               GREEN_LED_ON(); // Green LED on when USB connection established
            else
               GREEN_LED_OFF(); // Green LED off when no USB connection
            break;
         case 1:
         case 2:
            break;
         case 3:
            if (usbActivityFlag) { // Greeen LED flashes on USB activity
               GREEN_LED_TOGGLE();
               usbActivityFlag = 0;
            }
            break;            
         default :
            flashCount = 0;
            break;
         }
   }
}
#pragma MESSAGE DEFAULT C4003

//==================================================================
// Handler for USB Bus reset
// 
void handleUSBReset(void) {
   ERRSTAT = 0xFF;                 // Clear USB error flags
   // Clear all USB interrupt flags
   INTSTAT = INTSTAT_STALLF_MASK|INTSTAT_RESUMEF_MASK|INTSTAT_SLEEPF_MASK| 
             INTSTAT_TOKDNEF_MASK|INTSTAT_SOFTOKF_MASK|INTSTAT_ERRORF_MASK|INTSTAT_USBRSTF_MASK;

   ERRENB  = 0x00;                 // Disable all USB error interrupt sources

   // Enable various interrupts
   INTENB  = USB_INTMASKS;
   
   setUSBdefaultState();
}

//==================================================================
// Handler for USB Suspend
//
// * Enables the USB module to wakeup the CPU 
// * Stops the CPU
// On wakeup
// * Re-checks the USB bus after a small delay to avoid wakeups by noise
//
void handleUSBSuspend(void) {

   INTSTAT_SLEEPF    = 1;   // Clear the sleep int flag                    
   // Need to disable BDM interface & everything else to reduce power
   bdm_suspend();
   deviceState.state = USBsuspended;

   // A re-check loop is used here to ensure USB bus noise doesn't wakeup the CPU
   for(;;) {
      INTSTAT = INTSTAT_RESUMEF_MASK;  // Clear resume int flag
      INTENB_RESUME     = 1;           // Enable resume detection interrupts on the USB bus
      USBCTL0_USBRESMEN = 1;           // Allow the USB module to wakeup the CPU
     
      asm ("stop");                    // Processor stop for low power
      // The CPU has woken up!

      INTSTAT = INTSTAT_RESUMEF_MASK;  // Clear resume int flag
      asm { // Wait 3 us
         ldhx     #(3*BUS_FREQ)/(4*1000)
      Loop:
         dbnzx    Loop  ; [4]
      }
      // We should have another resume int by now
      if (INTSTAT&INTSTAT_RESUMEF_MASK)
         break;
   }
   USBCTL0_USBRESMEN = 0;  

return;
}

//==================================================================
// Handler for USB Resume
// 
// Disables further USB module wakeups
void handleUSBResume(void){
   INTENB_RESUME = 0;       // Mask further resume ints
   CTL_TSUSPEND  = 0;	    // Enable the transmit or receive of packets
   deviceState.state = USBconfigured;
   // power up BDM interface?
}

//==================================================================
// Handler for USB interrupt
// 
// Determines source and dispatches to appropriate routine.
//
interrupt //VectorNumber_Vusb 
void USBInterruptHandler( void ) {
U8 interruptFlags = INTSTAT;

   if((USBCTL0_LPRESF) && (deviceState.state==USBsuspended)) {
      USBCTL0_USBRESMEN = 0;
   }

   if ((interruptFlags&INTSTAT_RESUMEF_MASK) != 0) { // Resume signalled on Bus?
      handleUSBResume();
      INTSTAT = INTSTAT_RESUMEF_MASK; // Clear source
   }
   else if ((interruptFlags&INTSTAT_USBRSTF_MASK) != 0) {
      handleUSBReset();
      INTSTAT = INTSTAT_USBRSTF_MASK; // Clear source
   }
   else if ((interruptFlags&INTSTAT_TOKDNEF_MASK) != 0) { // Token complete int?
      U8 status = STAT;               // Snapshot status register
      INTSTAT = INTSTAT_TOKDNEF_MASK; // Clear source
      handleTokenComplete(status);
      usbActivityFlag = 1;
   }
   else if ((interruptFlags&INTSTAT_STALLF_MASK) != 0) { // Stall sent?
      handleEp0StallComplete();
      INTSTAT = INTSTAT_STALLF_MASK; // Clear source
   }
   else if ((interruptFlags&INTSTAT_SOFTOKF_MASK) != 0) { // SOF Token?
      handleSOFToken();
      INTSTAT = INTSTAT_SOFTOKF_MASK; // Clear source
   }
   else if ((interruptFlags&INTSTAT_SLEEPF_MASK) != 0) { // Bus Idle 3ms? => sleep
      handleUSBSuspend();
      INTSTAT = INTSTAT_SLEEPF_MASK; // Clear source
   }
   else  // unexpected int
      INTSTAT = interruptFlags; // Clear & ignore
}

#ifdef USE_USBPT1
void receiveCommand(int *size, U8 *buffer){

   ep2StartOutTransaction( *size, buffer ); 
     
   while (ep2State.state != EPDataAvailable) {
   }
   
   *size = (U8)ep2State.dataCount;
   
   ep2State.state = EPIdle;         
}

void sendResponse( int size, U8 *buffer){

   ep1StartInTransaction( size, buffer ); 
}
#endif